
Object subclass: #SkipList
  instanceVariableNames: 
    ' height probability header trailer '
  classVariableNames: ''
  poolDictionaries: ''
  categories: #( 'JOOP3' )

!

Object subclass: #SkipListElement
  instanceVariableNames: 
    ' key value nexts '
  classVariableNames: ''
  poolDictionaries: ''
  categories: #( 'JOOP3' )

!

SkipListElement subclass: #SkipListHeader
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: ''
  categories: #( 'JOOP3' )

!

SkipListElement subclass: #SkipListTrailer
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: ''
  categories: #( 'JOOP3' )

!


!CodeFiler forClass: SkipList class !
 
comment
    "Changed 'level' to 'height'. Replaced header by an ordered collection to eliminate
    'maximumLevel'. Eliminate direct access to skip list element's nexts; provide
    'nextAt:' and 'nextAt:put:'."! !

(CodeFiler organizerFor: SkipList class) reorganizeFrom: #(
('comment' comment )) !

!CodeFiler forClass: SkipList class !
  
probability: aFloat maximumHeight: anInteger
    ^self new probability: aFloat maximumHeight: anInteger!
   
new: initialSize
    "See Bill Pugh, Dept of Computer Science, Univ. of Maryland, paper at ftp://ftp.cs.umd.edu/pub/skipLists"
    | initialProbability initialHeight |
    initialProbability := 0.25.
    initialHeight := (initialSize log: 10) / ((1.0 / initialProbability) log: 10).
    ^self
        probability: initialProbability
        maximumHeight: (1 max: initialHeight ceiling)! !

(CodeFiler organizerFor: SkipList class) reorganizeFrom: #(
('creation' probability:maximumHeight: new: )) !

!CodeFiler forClass: SkipList class !
  
example1
    "SkipList example1"
    "SkipList example1 search: 50"
    "SkipList example1 remove: 50"
    | list |
    list := SkipList new: 50.
    #(10 50 20 40 5 100) do: [:integer |
        list insert: integer value: integer printString].
    ^list!
 
example2
    "SkipList example2"
    "SkipList example2 at: 50"
    "SkipList example2 at: 50 put: 'Smile'; yourself"
    | list |
    list := SkipList new: 50.
    #('seven' 'two' 'four' 'one' 'five' 'three') do: [:count |
        list at: count put: count size].
    ^list! !

(CodeFiler organizerFor: SkipList class) reorganizeFrom: #(
('examples' example1 example2 )) !


!CodeFiler forClass: SkipList !
   
probability: aFloat maximumHeight: anInteger
    header := SkipListHeader height: anInteger.
    trailer := SkipListTrailer height: anInteger.
    1 to: anInteger do: [:height | header nextAt: height put: trailer].
    self probability: aFloat; height: 1 "of empty list"! !

(CodeFiler organizerFor: SkipList) reorganizeFrom: #(
('initializing' probability:maximumHeight: )) !

!CodeFiler forClass: SkipList !
  
probability: aFloat
    probability := aFloat!
 
probability
    ^probability!
  
height: anInteger
    height := anInteger!
 
height
    ^height! !

(CodeFiler organizerFor: SkipList) reorganizeFrom: #(
('get/set' height height: probability: probability )) !

!CodeFiler forClass: SkipList !
 
maximumHeight
    ^header height
! !

(CodeFiler organizerFor: SkipList) reorganizeFrom: #(
('querying' maximumHeight )) !

!CodeFiler forClass: SkipList !
  
atHeight: height do: aBlock
    self atHeight: height listElementDo: [:element |
        aBlock value: element value]!

atHeight: height keysAndValuesDo: aBlock
    self atHeight: height listElementDo: [:element |
        aBlock value: element key value: element value]!

keysAndValuesDo: aBlock
    self atHeight: 1 keysAndValuesDo: aBlock!
  
atHeight: height listElementDo: aBlock
    "Sequence through the list at the given height and supply the skip
    list elements (exclude the header/trailer)."
    | current next |
    current := header nextAt: height.
    next := current nextAt: height.
    [next notNil] whileTrue: [
        aBlock value: current.
        current := next. next := current nextAt: height].!
  
do: aBlock
    self atHeight: 1 do: aBlock! !

(CodeFiler organizerFor: SkipList) reorganizeFrom: #(
('sequencing' do: atHeight:keysAndValuesDo: atHeight:do: atHeight:listElementDo: keysAndValuesDo: )) !

!CodeFiler forClass: SkipList !
  
search: aMagnitude
    ^self search: aMagnitude ifAbsent: [^nil]!
  
insert: aMagnitude value: anObject
    | predecessors predecessor next element |
    predecessors := self predecessorsOf: aMagnitude.
    next := predecessors first next.
    (next = aMagnitude) ifTrue: [next value: anObject. ^self].

    "Insert element after predecessors elements."
    element := SkipListElement
        height: self randomHeight
        key: aMagnitude
        value: anObject.
    1 to: element height do: [:height |
        predecessor := predecessors at: height.
        predecessor atHeight: height insertNext: element].
    self height: (self height max: element height)
!
   
search: aMagnitude ifAbsent: aBlock
    | element |
    element := self atOrPast: aMagnitude.
    element = aMagnitude
        ifTrue: [^element value]
        ifFalse: [^aBlock value].!
 
remove: aMagnitude
    self remove: aMagnitude ifAbsent: [self error: 'key not found']!

remove: aMagnitude ifAbsent: aBlock
    | predecessors predecessor element |
    predecessors := self predecessorsOf: aMagnitude.
    element := predecessors first next.
    element = aMagnitude ifFalse: [^aBlock value].

    "Link around this element."
    1 to: element height do: [:height |
        predecessor := predecessors at: height.
        predecessor nextAt: height put: (element nextAt: height)].

    "Eliminate unnecessary levels."
    self height to: 2 by: -1 do: [:index |
        (header nextAt: index) == trailer ifFalse: [^self].
        self height: index]! !

(CodeFiler organizerFor: SkipList) reorganizeFrom: #(
('searching/inserting/deleting' remove:ifAbsent: search: insert:value: search:ifAbsent: remove: )) !

!CodeFiler forClass: SkipList !
 
atOrPast: aMagnitude
    "The first skip list element whose key is >= aMagnitude."
    | previous next |
    "Move right at the top level as far as possible, then drop down and try
    to go further. Retain the previous skip list element at each level."
    previous := header.
    self height to: 1 by: -1 do: [:height |
        next := previous nextAt: height.
        [next < aMagnitude "i.e., aMagnitude < next"] whileTrue: [
            previous := next. next := previous nextAt: height].
        "Drop down one level and go further."].
    "Return the last element at height 1."
   ^next!
  
predecessorsOf: aMagnitude
    "All skip list elements for each height (header allowed)."
    | predecessors previous next |
    "Move right at the top level as far as possible, then drop down and try
    to go further. Retain the previous skip list element at each level."
    predecessors := (Array new: self maximumHeight) atAllPut: header.
    previous := header.
    self height to: 1 by: -1 do: [:height |
        next := previous nextAt: height.
        [next < aMagnitude "i.e., aMagnitude < next"] whileTrue: [
            previous := next. next := previous nextAt: height].
        "Save the previous element."
        predecessors at: height put: previous.
        "Drop down one level and go further."].
    "Return the previous element at all levels."
   ^predecessors! !

(CodeFiler organizerFor: SkipList) reorganizeFrom: #(
('searching support' atOrPast: predecessorsOf: )) !

!CodeFiler forClass: SkipList !

at: key put: value
    self insert: key value: value!
  
at: key ifAbsent: aBlock
    ^self search: key ifAbsent: aBlock!
   
at: key
    ^self at: key ifAbsent: [self error: 'key not found']! !

(CodeFiler organizerFor: SkipList) reorganizeFrom: #(
('dictionary access' at:put: at:ifAbsent: at: )) !

!CodeFiler forClass: SkipList !
   
randomHeight
    | myProbability |
    myProbability := (100.0 * self probability) truncated.
    1 to: self maximumHeight do: [:index |
        (Random from0To: 99) >= myProbability
            ifTrue: [^index]].
    ^self maximumHeight! !

(CodeFiler organizerFor: SkipList) reorganizeFrom: #(
('random' randomHeight )) !

!CodeFiler forClass: SkipList !
 
printOn: aStream
    self maximumHeight to: 1 by: -1 do: [:height |
        aStream nextPutAll: '('; print: height; nextPutAll: ':'.
        self atHeight: height listElementDo: [:element |
            aStream space; print: element].
        aStream nextPutAll: ') '.]! !

(CodeFiler organizerFor: SkipList) reorganizeFrom: #(
('printing' printOn: )) !
CodeFiler setCommentFor: SkipList to: ''!


!CodeFiler forClass: SkipListElement class !
   
height: anInteger key: aMagnitude value: anObject
    ^self new height: anInteger key: aMagnitude value: anObject!
 
height: anInteger
    ^self height: anInteger key: nil value: nil! !

(CodeFiler organizerFor: SkipListElement class) reorganizeFrom: #(
('creation' height:key:value: height: )) !


!CodeFiler forClass: SkipListElement !
 
height: anInteger key: aMagnitude value: anObject
    nexts := Array new: anInteger.
    self key: aMagnitude; value: anObject! !

(CodeFiler organizerFor: SkipListElement) reorganizeFrom: #(
('initializing' height:key:value: )) !

!CodeFiler forClass: SkipListElement !
   
key: aMagnitude
    key := aMagnitude!
 
key
    ^key!
  
value: anObject
    value := anObject!
 
value
    ^value! !

(CodeFiler organizerFor: SkipListElement) reorganizeFrom: #(
('get/set' value key: key value: )) !

!CodeFiler forClass: SkipListElement !
   
height
    ^nexts size! !

(CodeFiler organizerFor: SkipListElement) reorganizeFrom: #(
('querying' height )) !

!CodeFiler forClass: SkipListElement !
   
nextAt: height
    ^nexts at: height!
  
nextAt: height put: aSkipListElement
    ^nexts at: height put: aSkipListElement!
  
next
    ^self nextAt: 1! !

(CodeFiler organizerFor: SkipListElement) reorganizeFrom: #(
('next facilities' nextAt: nextAt:put: next )) !

!CodeFiler forClass: SkipListElement !

atHeight: height insertNext: aSkipListElement
    | after |
    after := self nextAt: height.
    self nextAt: height put: aSkipListElement.
    aSkipListElement nextAt: height put: after! !

(CodeFiler organizerFor: SkipListElement) reorganizeFrom: #(
('linking' atHeight:insertNext: )) !

!CodeFiler forClass: SkipListElement !
  
< aMagnitude
    ^self key < aMagnitude!
   
= aMagnitude
    ^self key = aMagnitude! !

(CodeFiler organizerFor: SkipListElement) reorganizeFrom: #(
('comparing (private)' < = )) !

!CodeFiler forClass: SkipListElement !
  
printOn: aStream
    aStream print: self key; nextPutAll: '=>'; print: self value! !

(CodeFiler organizerFor: SkipListElement) reorganizeFrom: #(
('printing' printOn: )) !
CodeFiler setCommentFor: SkipListElement to: ''!



!CodeFiler forClass: SkipListHeader !
 
< aMagnitude
    ^true!

= aMagnitude
    ^false! !

(CodeFiler organizerFor: SkipListHeader) reorganizeFrom: #(
('comparing (private)' < = )) !

!CodeFiler forClass: SkipListHeader !

printOn: aStream
    aStream nextPutAll: 'header'! !

(CodeFiler organizerFor: SkipListHeader) reorganizeFrom: #(
('printing' printOn: )) !
CodeFiler setCommentFor: SkipListHeader to: ''!



!CodeFiler forClass: SkipListTrailer !
  
< aMagnitude
    ^false!
   
= aMagnitude
    ^false! !

(CodeFiler organizerFor: SkipListTrailer) reorganizeFrom: #(
('comparing (private)' < = )) !

!CodeFiler forClass: SkipListTrailer !
  
printOn: aStream
    aStream nextPutAll: 'trailer'! !

(CodeFiler organizerFor: SkipListTrailer) reorganizeFrom: #(
('printing' printOn: )) !
CodeFiler setCommentFor: SkipListTrailer to: ''!
