
Object subclass: #SkipListElementVersion1
  instanceVariableNames: 
    ' key value nexts '
  classVariableNames: ''
  poolDictionaries: ''
  categories: #( 'JOOP3' )

!

Object subclass: #SkipListVersion1
  instanceVariableNames: 
    ' level maximumLevel probability header '
  classVariableNames: ''
  poolDictionaries: ''
  categories: #( 'JOOP3' )

!


!CodeFiler forClass: SkipListElementVersion1 class !
 
level: anInteger key: aMagnitude value: anObject
    ^self new level: anInteger key: aMagnitude value: anObject! !

(CodeFiler organizerFor: SkipListElementVersion1 class) reorganizeFrom: #(
('creation' level:key:value: )) !


!CodeFiler forClass: SkipListElementVersion1 !

level: anInteger key: aMagnitude value: anObject
    self nexts: (Array new: anInteger); key: aMagnitude; value: anObject! !

(CodeFiler organizerFor: SkipListElementVersion1) reorganizeFrom: #(
('initializing' level:key:value: )) !

!CodeFiler forClass: SkipListElementVersion1 !
  
nexts: skipListElementCollection
    nexts := skipListElementCollection!
   
key: aMagnitude
    key := aMagnitude!
 
key
    ^key!
  
nexts
    ^nexts!
  
value: anObject
    value := anObject!
 
value
    ^value! !

(CodeFiler organizerFor: SkipListElementVersion1) reorganizeFrom: #(
('get/set' nexts: key: key nexts value: value )) !

!CodeFiler forClass: SkipListElementVersion1 !
  
printOn: aStream
    aStream print: self key; nextPutAll: '=>'; print: self value! !

(CodeFiler organizerFor: SkipListElementVersion1) reorganizeFrom: #(
('printing' printOn: )) !
CodeFiler setCommentFor: SkipListElementVersion1 to: ''!


!CodeFiler forClass: SkipListVersion1 class !
   
comment
    "Same implementation as in 'Thomas Wenger, The Elegant (and Fast) Skip List, Java Pro,
    April/May 98'. But modified so that level is a number >= 1 rather than a number >= 0.
    Also, abstracted search methods for use by search, insert, and delete methods and
    added a printOn: method."! !

(CodeFiler organizerFor: SkipListVersion1 class) reorganizeFrom: #(
('comment' comment )) !

!CodeFiler forClass: SkipListVersion1 class !

new: initialSize
    "See Bill Pugh, Dept of Computer Science, Univ. of Maryland, paper at ftp://ftp.cs.umd.edu/pub/skipLists"
    | initialProbability initialLevel |
    initialProbability := 0.25.
    initialLevel := (initialSize log: 10) / ((1.0 / initialProbability) log: 10).
    ^self
        probability: initialProbability
        maximumLevel: (1 max: initialLevel ceiling)!
  
probability: aFloat maximumLevel: anInteger
    ^self new probability: aFloat maximumLevel: anInteger! !

(CodeFiler organizerFor: SkipListVersion1 class) reorganizeFrom: #(
('creation' new: probability:maximumLevel: )) !

!CodeFiler forClass: SkipListVersion1 class !
  
example1
    "SkipListVersion1 example1"
    "SkipListVersion1 example1 search: 50"
    "SkipListVersion1 example1 remove: 50"
    | list |
    list := SkipListVersion1 new: 50.
    #(10 50 20 40 5 100) do: [:integer |
        list insert: integer value: integer printString].
    ^list! !

(CodeFiler organizerFor: SkipListVersion1 class) reorganizeFrom: #(
('examples' example1 )) !


!CodeFiler forClass: SkipListVersion1 !

probability: aFloat maximumLevel: anInteger
    | header trailer |
    header := SkipListElementVersion1 level: anInteger key: 0 value: nil.
    trailer := SkipListElementVersion1 level: anInteger key: 10000000000 value: nil.
    self
        probability: aFloat;
        level: 1;
        maximumLevel: anInteger;
        header: header.
    1 to: self maximumLevel do: [:currentLevel |
        header nexts at: currentLevel put: trailer]
! !

(CodeFiler organizerFor: SkipListVersion1) reorganizeFrom: #(
('initializing' probability:maximumLevel: )) !

!CodeFiler forClass: SkipListVersion1 !
  
header: aSkipListElementVersion1
    header := aSkipListElementVersion1!
   
maximumLevel: anInteger
    maximumLevel := anInteger!
 
probability: aFloat
    probability := aFloat!
 
probability
    ^probability!
  
level: anInteger
    level := anInteger!
   
level
    ^level!
  
maximumLevel
    ^maximumLevel!

header
    ^header! !

(CodeFiler organizerFor: SkipListVersion1) reorganizeFrom: #(
('get/set' header: maximumLevel: probability: probability level: level maximumLevel header )) !

!CodeFiler forClass: SkipListVersion1 !
 
atOrPast: aMagnitude
    "The first skip list element whose key is >= aMagnitude."
    | cursor next |
    "Move right at the top level as far as possible, then drop down and try
    to go further."
    cursor := self header.
    self level to: 1 by: -1 do: [:currentLevel |
        next := cursor nexts at: currentLevel.
        [aMagnitude > next key] whileTrue: [
            cursor := next. next := cursor nexts at: currentLevel].
        "Drop down one level and go further."].
   ^next!
 
search: aMagnitude
    | element |
    element := self atOrPast: aMagnitude.
    element key = aMagnitude
        ifTrue: [^element value]
        ifFalse: [^nil].! !

(CodeFiler organizerFor: SkipListVersion1) reorganizeFrom: #(
('searching' atOrPast: search: )) !

!CodeFiler forClass: SkipListVersion1 !

insert: aMagnitude value: anObject
    | predecessors element newLevel after |
    predecessors := self predecessorsOf: aMagnitude.
    element := predecessors first nexts at: 1.
    (element key = aMagnitude) ifTrue: [
        element value: anObject. ^self].

    "Obtain a random level for insertion."
    newLevel := self randomLevel.
    newLevel > self level ifTrue: [
        self level + 1 to: newLevel do: [:index |
            predecessors at: index put: self header].
        self level: newLevel].

    "Insert element after predecessors elements."
    element := SkipListElementVersion1 level: newLevel key: aMagnitude value: anObject.
    1 to: newLevel do: [:index |
        after := (predecessors at: index) nexts at: index.
        (predecessors at: index) nexts at: index put: element.
        element nexts at: index put: after]
!
   
predecessorsOf: aMagnitude
    "All skip list elements for each height (header allowed)."
    | predecessors cursor next |
    "Move right at the top level as far as possible, then drop down and try
    to go further. Retain the next skip list element at each level"
    predecessors := Array new: self maximumLevel.
    cursor := self header.
    self level to: 1 by: -1 do: [:currentLevel |
        next := cursor nexts at: currentLevel.
        [aMagnitude > next key] whileTrue: [
            cursor := next. next := cursor nexts at: currentLevel].
        "Save the previous element."
        predecessors at: currentLevel put: cursor.
        "Drop down one level and go further."].
    "Return the previous element at all levels."
   ^predecessors! !

(CodeFiler organizerFor: SkipListVersion1) reorganizeFrom: #(
('inserting' insert:value: predecessorsOf: )) !

!CodeFiler forClass: SkipListVersion1 !
   
remove: aMagnitude
    | predecessors element newLevel after |
    predecessors := self predecessorsOf: aMagnitude.
    element := predecessors first nexts at: 1.
    element key = aMagnitude ifFalse: [
        self error: 'key not found'].

    "Link around this element."
    1 to: self level do: [:index |
        after := (predecessors at: index) nexts at: index.
        after == element ifTrue: [
            (predecessors at: index) nexts at: index put: (element nexts at: index)]].

    "Eliminate unnecessary levels."
    self level to: 2 by: -1 do: [:index |
        (self header nexts at: index) key = 10000000000 ifFalse: [^self].
        self level: index]! !

(CodeFiler organizerFor: SkipListVersion1) reorganizeFrom: #(
('deleting' remove: )) !

!CodeFiler forClass: SkipListVersion1 !
  
randomLevel
    | myProbability |
    myProbability := (100.0 * self probability) truncated.
    1 to: self maximumLevel do: [:index |
        (Random from0To: 99) >= myProbability
            ifTrue: [^index]].
    ^self maximumLevel! !

(CodeFiler organizerFor: SkipListVersion1) reorganizeFrom: #(
('random' randomLevel )) !

!CodeFiler forClass: SkipListVersion1 !
 
printOn: aStream
    | cursor |
    self maximumLevel to: 1 by: -1 do: [:currentLevel |
        aStream nextPutAll: '('; print: currentLevel; nextPutAll: ':'.
        cursor := self header.
        [cursor notNil] whileTrue: [
            aStream space; print: cursor.
            cursor := cursor nexts at: currentLevel].
        aStream nextPutAll: ') '.]
! !

(CodeFiler organizerFor: SkipListVersion1) reorganizeFrom: #(
('printing' printOn: )) !
CodeFiler setCommentFor: SkipListVersion1 to: ''!
